/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.semantic.document;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import it.tidalwave.util.NotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import lombok.RequiredArgsConstructor;
import static it.tidalwave.semantic.document.Document.*;

/***********************************************************************************************************************
 *
 * A {@link DocumentVisitor} which creates a zip file containing the attachments of a {@link Document}.
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@NotThreadSafe @RequiredArgsConstructor
public class AttachmentZipGenerator extends DocumentVisitorSupport<IOException>
  {
    private static final Logger log = LoggerFactory.getLogger(AttachmentZipGenerator.class);
    
    private final @Nonnull File file;
    
    @CheckForNull
    private ZipOutputStream zos;
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void preVisitBody (final @Nonnull Document document) 
      throws FileNotFoundException 
      {
        zos = new ZipOutputStream(new FileOutputStream(file));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void postVisitBody (final @Nonnull Document document)
      throws IOException 
      {
        zos.finish();
        zos.flush();
        zos.close();        
        log.debug(">>>> Attachment saved to {}, {} bytes", file, file.length());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void visitAttachment (final @Nonnull Document attachment) 
      throws IOException 
      {
        final String subject;
        try 
          {
            subject = attachment.get(SUBJECT);
          }
        catch (NotFoundException e)
          {
            throw new IOException(e.toString()); // Java 5 compatibility
          }
        
        log.debug(">>>> storing attachment: {}", subject);
        final ZipEntry zipEntry = new ZipEntry(subject);
        zos.putNextEntry(zipEntry);
        
        try 
          {
            zos.write(attachment.get(BODY).getBytes());
          }
        catch (NotFoundException e) 
          {
            // ok, no body
          }
        
        zos.closeEntry();
      }    
  }
